//------------------------------------------------------------------------------
// Copyright (C) 2010, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <glib.h>

#include <stdio.h>
#include <mbus.h>

#include "net.h"

static int address = 0;

static gboolean server_io(GIOChannel *gio, GIOCondition condition, gpointer data);
static GIOChannel *gio_server;
#define PACKET_BUFF_SIZE 128

//------------------------------------------------------------------------------
// send a data request packet to from master to slave
//------------------------------------------------------------------------------
static gboolean
send_packet_timeout(gpointer data)
{
    static gchar *func_name = "send_packet_timer";
    char buff[256];
    int len;

    mbus_frame *frame;
    
    address++;
    
    if (address > 250)
    {
        exit(0);
    }    
    
    frame = mbus_frame_new(MBUS_FRAME_TYPE_SHORT);    
    frame->control  = MBUS_CONTROL_MASK_SND_NKE | MBUS_CONTROL_MASK_DIR_M2S;
    frame->address  = address; 

    //if ((ret = mbus_frame_print(frame)) < 0)
    //{
    //    printf("%s: mbus_frame_print -> %d\n", __PRETTY_FUNCTION__, ret);
    //    return -1;
    //}
   
    if ((len = mbus_frame_pack(frame, buff, sizeof(buff))) == -1)
    {
        printf("%s: mbus_frame_pack -> %d\n", __PRETTY_FUNCTION__, len);
        return -1;
    }    

    if (fs_net_send(gio_server, buff, len) < len)
    { 
        g_message("%s: failed to write data to server", func_name);
    }    
	
    return FALSE; 
}

//------------------------------------------------------------------------------
// client_io:
//
//------------------------------------------------------------------------------
static gboolean
server_io(GIOChannel *gio, GIOCondition condition, gpointer data)
{
    static gchar *func_name = "server_io";
    gchar buff[PACKET_BUFF_SIZE];
    gint len;
    int fd;
    FsNetBuffer *nbdata;

    bzero((void *)buff, sizeof(buff));

    mbus_frame *frame;

    //
    // init
    //

    if ((fd = g_io_channel_unix_get_fd(gio)) < 0)
    {
       g_warning("%s: Failed to aquire socket fd.", func_name);
       return FALSE;
    }

    if (condition & (G_IO_HUP|G_IO_NVAL|G_IO_ERR))
    {
        exit(0);
    }

    if ((nbdata  = fs_net_buffer_get(gio)) == NULL)
    {
        g_warning("%s: Serious error, couldn't retrieve data buffer for GIO Channel.", func_name);
        return FALSE;
    }

    //
    // start reading/parsing data
    //
    frame = mbus_frame_new(MBUS_FRAME_TYPE_ANY);
    len = 1;
    
    while (len > 0)
    {
        if ((len = fs_net_read_data_nonblocking(gio, nbdata, len)) < 0)
        {
            if (len == -2 && nbdata->offset == 0)
            {
                // Return value -2 is a special return value that we use to indicate
                // that the read operation was interrupted due to unavailability of 
                // data (so we got an try-again error, preventing a blocking read).
                //
                // Restart the read operation later.
                return TRUE;
            }        
	
	        if (len == -1)
	        {
	            g_message("%s: Closing socket, out of data (%d).", func_name, fd);
                fs_net_buffer_free(gio);
	            fs_net_socket_close(fd);
                exit(0);
            }
        }
        
        nbdata->offset = len; // start next read at the correct offset
        len = mbus_parse(frame, nbdata->buff, len);       
    }
  
    if (len == -1)
    {
        g_message("%s: M-Bus layer failed to parse data", func_name);
        mbus_frame_free(frame);
        return TRUE;
    }
  
    //mbus_frame_print(frame);

    if (mbus_frame_type(frame) == MBUS_FRAME_TYPE_ACK)
    {
        printf("Found a M-Bus device at address %d\n", address);
    }

    mbus_frame_free(frame);

    return TRUE;
}

//------------------------------------------------------------------------------
// Execution starts here:
//
//------------------------------------------------------------------------------
int
main(int argc, char **argv)
{
    static gchar *func_name = "mbus_test1", *host;
    int port;

    GMainLoop *gloop;
 
    if (argc != 3)
    {
        g_message("%s: usage: %s host port", func_name, argv[0]);
        return 0;
    }
 
    host = argv[1];   
    port = atoi(argv[2]);
    
    if ((gio_server = fs_net_connection_setup(host, port, (void *)server_io, NULL)) == NULL)
    {
        g_warning("%s: Failed to setup connection to server.\n", func_name);
        return -1;
    }

    // Register new data buffer for the new GIO Channel.
    if (fs_net_buffer_new(gio_server) != 0)
    {
        g_warning("%s: Failed to allocate GIO channel data buffer.", __PRETTY_FUNCTION__);
        return -1;
    }
   	
    /* setup a timer */
    g_timeout_add(100, send_packet_timeout, NULL);
   	
    /* Pass over flow control to GLIB */
    gloop = g_main_loop_new(NULL, FALSE);
    g_main_loop_run(gloop);

    /* Never reached */
    return 0;
}



